Otis = LibStub("AceAddon-3.0"):NewAddon("Otis", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Otis")

--==== set upvalues ====--
-- GLOBALS: LibStub, BonusRollFrame, EncounterJournal, EncounterJournalEncounterFrameInfoLootTab, EncounterJournalEncounterFrameInfoModelTab, EncounterJournal_DisplayEncounter, EncounterJournal_DisplayInstance
local _G, Otis, pairs, rawset, select, type, CreateFrame, GetLootSpecialization, GetNumSpecializationsForClassID, GetSpecializationInfo, SetLootSpecialization, UnitClass, EJ_GetCreatureInfo, EJ_GetEncounterInfo, EJ_GetEncounterInfoByIndex, EJ_GetLootFilter = 
      _G, Otis, pairs, rawset, select, type, CreateFrame, GetLootSpecialization, GetNumSpecializationsForClassID, GetSpecializationInfo, SetLootSpecialization, UnitClass, EJ_GetCreatureInfo, EJ_GetEncounterInfo, EJ_GetEncounterInfoByIndex, EJ_GetLootFilter


Otis.DL = {
	[0] = "silent",
	"error",
	"standard",
	"test",
	"config",
	"events",
	"spam",
}
local DL = Otis.DL
do	-- add lookups for name to index and index as string to index to table  -- review - shouldn't it be index as string to name? check other addons
	for i=0, #DL do
		DL[ DL[i] ] = i
		DL[tostring(i)] = i
	end
end

local dbgprint = { print   = function(...) return Otis:Print(...) end,
                   squelch = function(...) return nil end }
function Otis:SetDebugLevel(iDL)
	for i = 0,     iDL do dbgprint[ DL[i] ] = dbgprint.print;   end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end


-- iterator function to return boss names from a given encounter
local function EncounterBosses(encounter, index)
	index = (index or 0) + 1
	local _, name = EJ_GetCreatureInfo(index, encounter)
	if name then
		return index, name
	end
end

-- iterator function to return encounter ids
local function InstanceEncounters(invariant, index)
	index = (index or 0) + 1
	local _, _, encounter = EJ_GetEncounterInfoByIndex(index)
	if encounter then
		-- dbgprint.spam(index, name, encounter)
		return index, encounter
	end
end


local function GetEncounters(t,k)
	local info = { ByIndex = {}, ByID = {} }
	for index, encounter in InstanceEncounters do
		info.ByIndex[index] = encounter
		info.ByID[encounter] = index
	end

	rawset(t, k, info)
	return info
end

local g_tblInstances = setmetatable( {}, { __index = GetEncounters } )
local g_tblSpecNames = { [0] = "Active" }
local g_tblSpecTex = {}
local g_idPlayerClass
local g_idPendingSpec
local g_fWaitingOnBonusLoot = false
local g_idActiveDifficulty

local dbbosses
local dbworld
local dbchar

local TimerRestore = { func = "RestoreLootSpec", delay = 5 }
function TimerRestore:Set(event)
	if self.timer then
		Otis:CancelTimer(self.timer)
	end
	self.timer = Otis:ScheduleTimer(self.func, self.delay)
	dbgprint.spam("scheduling restore for event", event)
end
function TimerRestore:Cancel()
	if self.timer then
		Otis:CancelTimer(self.timer)
		self.timer = nil
	end
end


-- calling this with nil or 0 as the spec ID should hide all icons
local function ShowSpec(specID)
	for id, tex in pairs(g_tblSpecTex) do
		if id == specID then
			tex:Show()
		else
			tex:Hide()
		end
	end
end

function Otis:UpdateLootSpecButton()
	local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID
	local difficultyID = dbworld[encounterID] and 0 or EJ_GetDifficulty()
	-- dbgprint.test("loot spec button update", instanceID, encounterID, difficultyID)
	if encounterID then
		ShowSpec(dbchar.encounters[encounterID][difficultyID])
		self.btnLootSpec:Enable()
	else
		ShowSpec(nil)
		self.btnLootSpec:Disable()
	end
end
	
local function btnLootSpec_Click(self, button, down)
	local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID

	if type(encounterID) == "number" then 
		if button == "LeftButton" then
			-- save the spec pref
			local classID, specID = EJ_GetLootFilter()
			
			if g_idPlayerClass ~= classID then
				dbgprint.error(L["ERROR: wrong class selected"])
			elseif specID == 0 then
				dbgprint.error(L["ERROR: no spec selected"])
			else
				-- change encounters table
				if dbworld[encounterID] then
					dbchar.encounters[encounterID][0] = specID
				elseif IsShiftKeyDown() then
					for difficultyID = 1, 17 do
						if EJ_IsValidInstanceDifficulty(difficultyID) then
							dbchar.encounters[encounterID][difficultyID] = specID
						end
					end
				else
					local difficultyID = EJ_GetDifficulty()
					if difficultyID and EJ_IsValidInstanceDifficulty(difficultyID) then
						dbchar.encounters[encounterID][difficultyID] = specID
					end
				end

				-- change bosses table
				local index = 1
				for _, name in EncounterBosses, encounterID do
					dbbosses[name] = encounterID
				end

				-- change button overlay icon
				ShowSpec(specID)
			end
		elseif button == "RightButton" then
			local difficultyID = dbworld[encounterID] and 0 or EJ_GetDifficulty()
			if type(difficultyID) == "number" then -- clear setting for this encounter
				-- clear encounters table
				dbchar.encounters[encounterID][difficultyID] = nil

				-- hide button overlay icon
				ShowSpec(nil)
			end
		end
	end
end

local function btnLootSpec_Enter(self)
	local encounterID = EncounterJournal.encounterID

	GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
	GameTooltip:SetText(L["loot specialization manager"]);

	if encounterID and dbworld[encounterID] then
		GameTooltip:AddLine(L["left click to save"], 1, 1, 1, 1);
		GameTooltip:AddLine(L["right click to clear"], 1, 1, 1, 1);
	else
		GameTooltip:AddLine(L["left click to save for current difficulty"], 1, 1, 1, 1);
		GameTooltip:AddLine(L["shift + left click to save for all difficulties"], 1, 1, 1, 1);
		GameTooltip:AddLine(L["right click to clear for current difficulty"], 1, 1, 1, 1);
	end

	GameTooltip:Show();
end

local function btnLootSpec_Leave(self)
	GameTooltip:Hide();
end

local function ScrollToBoss(index)
	local sf = _G.EncounterJournalEncounterFrameInfoBossesScrollFrame
	local btn = _G["EncounterJournalBossButton"..index]
	local diffBottom = sf:GetBottom() - btn:GetBottom()
	local diffTop = btn:GetTop() - sf:GetTop()

	if diffBottom > 0 then
		-- dbgprint.test(format("need to scroll the pane up by %i", diffBottom))
		sf:SetVerticalScroll(sf:GetVerticalScroll() + diffBottom)
	end
	if diffTop > 0 then
		-- dbgprint.test(format("need to scroll the pane down by %i", diffTop))
		sf:SetVerticalScroll(sf:GetVerticalScroll() - 10 - diffTop)
	end
end

local function btnUp_Click()
	if EncounterJournal then
		local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID
		if instanceID and encounterID then
			local info = g_tblInstances[instanceID]
			local index = info.ByID[encounterID]
			if not index then
				dbgprint.test(format("btnUp_Click - nil index for encounter %i in instance %i", encounterID, instanceID))
			elseif index > 1 then
				EncounterJournal_DisplayEncounter(info.ByIndex[index-1], true)
				ScrollToBoss(index-1)
			elseif index == 1 then
				EncounterJournal_DisplayInstance(instanceID, true)
			end
		end
	end
end

local function btnDown_Click()
	if EncounterJournal then
		local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID
		if instanceID then
			local info = g_tblInstances[instanceID]
			if encounterID then
				local index = info.ByID[encounterID]
				if not index then
					dbgprint.test(format("btnDown_Click - nil index for encounter %i in instance %i", encounterID, instanceID))
				elseif index < #info.ByIndex then
					EncounterJournal_DisplayEncounter(info.ByIndex[index+1], true)
					ScrollToBoss(index+1)
				end
			elseif #info.ByIndex > 0 then
				EncounterJournal_DisplayEncounter(info.ByIndex[1], true)
			end
		end
	end
end

function Otis:MakeButtons()
	if (type(EncounterJournalEncounterFrameInfoLootTab) ~= "table") then
		dbgprint.config("parent frame doesn't exist yet")
		return
	end

	dbgprint.config("making buttons")
	self:UnregisterEvent("EJ_LOOT_DATA_RECIEVED")
	self:UnregisterEvent("EJ_DIFFICULTY_UPDATE")

	local btnLootSpec = CreateFrame("Button", "OtisButtonLootSpec", EncounterJournalEncounterFrameInfoModelTab)
	self.btnLootSpec = btnLootSpec
	btnLootSpec:SetScript("OnClick", btnLootSpec_Click)
	btnLootSpec:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	btnLootSpec:SetScript("OnEnter", btnLootSpec_Enter)
	btnLootSpec:SetScript("OnLeave", btnLootSpec_Leave)

	btnLootSpec:SetPoint("TOP", EncounterJournalEncounterFrameInfoModelTab, "BOTTOM", 0, 2)
	btnLootSpec:SetSize(63, 57)
	btnLootSpec:SetNormalTexture("Interface\\EncounterJournal\\UI-EncounterJournalTextures")
	btnLootSpec:GetNormalTexture():SetTexCoord(0.25585938, 0.37890625, 0.90332031, 0.95898438)	-- UI-EJ-Tab-Selected
	btnLootSpec:SetDisabledTexture("Interface\\EncounterJournal\\UI-EncounterJournalTextures")
	btnLootSpec:GetDisabledTexture():SetTexCoord(0.25585938, 0.37890625, 0.90332031, 0.95898438)	-- UI-EJ-Tab-UnSelected
	btnLootSpec:GetDisabledTexture():SetDesaturated(true)
	btnLootSpec:SetHighlightTexture("Interface\\EncounterJournal\\UI-EncounterJournalTextures")
	btnLootSpec:GetHighlightTexture():SetTexCoord(0.00195313, 0.12500000, 0.90332031, 0.95898438)	-- UI-EJ-Tab-Highlight

	for i=1, GetNumSpecializationsForClassID(g_idPlayerClass) do
		local id, name, _, icon = GetSpecializationInfo(i)
		local tex = btnLootSpec:CreateTexture(name, "OVERLAY")
		tex:SetTexture(icon)
		tex:SetPoint("BOTTOMLEFT", 20, 13)
		tex:SetPoint("TOPRIGHT", -11, -12)
		tex:Hide()
		g_tblSpecTex[id] = tex
	end

	btnLootSpec:Disable()

 	-- self:SecureHook("EncounterJournal_ListInstances", "UpdateLootSpecButton")
	self:SecureHook("EncounterJournal_DisplayInstance",  "UpdateLootSpecButton")
	self:SecureHook("EncounterJournal_DisplayEncounter", "UpdateLootSpecButton")
	self:SecureHook("EncounterJournal_SelectDifficulty", "UpdateLootSpecButton")

	local btnUp = CreateFrame("Button", "OtisButtonUp", EncounterJournalEncounterFrameInfoModelTab)
	self.btnUp = btnUp
	btnUp:SetScript("OnClick", btnUp_Click)
--	btnUp:SetPoint("TOP", EncounterJournalEncounterFrameInfoModelTab, "BOTTOM", 6, 2)
	btnUp:SetPoint("TOP", btnLootSpec, "BOTTOM", 6, 2)
	btnUp:SetSize(63, 57)
	btnUp:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Up")
	btnUp:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Down")
	btnUp:SetDisabledTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Disabled")
	btnUp:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")

	local btnDown = CreateFrame("Button", "OtisButtonDown", EncounterJournalEncounterFrameInfoModelTab)
	self.btnDown = btnDown
	btnDown:SetScript("OnClick", btnDown_Click)
	btnDown:SetPoint("TOP", btnUp, "BOTTOM", 0, 2)
	btnDown:SetSize(63, 57)
	btnDown:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	btnDown:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down")
	btnDown:SetDisabledTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Disabled")
	btnDown:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")


--[==[
	btnUp:SetNormalTexture("Interface\\MainMenuBar\\UI-MainMenu-ScrollUpButton-Up")
	btnUp:SetPushedTexture("Interface\\MainMenuBar\\UI-MainMenu-ScrollUpButton-Down")
	btnUp:SetDisabledTexture("Interface\\Buttons\\UI-ScrollBar-ScrollUpButton-Disabled")
	btnUp:SetHighlightTexture("Interface\\MainMenuBar\\UI-MainMenu-ScrollUpButton-Highlight", "ADD")
--]==]
end


function Otis:LoadSavedSettings()
	-- saved settings and default values
	local defaults = {
		global = {
			bosses = {
				[ L["Mantid Spoils"] ] = 870,
				[ L["Mogu Spoils"] ] = 870,
			},
--[===[@debug@
			bossesSeen = {},
--@end-debug@]===]
			worldBosses = { -- strings do not need to be localized, we're only checking existence of keys in the table
				[691] = "Sha of Anger",
				[725] = "Saylis's Warband",
				[814] = "Nalak, The Storm Lord",
				[826] = "Oondasta",
				[857] = "Chi-ji, The Red Crane",
				[858] = "Yu'lon, The Jade Serpent",
				[859] = "Niuzao, The Black Ox",
				[860] = "Xuen, The White Tiger",
				[861] = "Ordos, Fire-God of the Yaungol",
				[1291] = "Drov the Ruiner",
				[1211] = "Tarina the Ageless",
				[1262] = "Ruhkmar",
				[1452] = "Supreme Lord Kazzak",
			},
			output = "standard",
		},
		char = {
			encounters = { ['*'] = {} },
		},
	}

	self.db = LibStub("AceDB-3.0"):New("OtisDB", defaults, true) --, _G.DEFAULT)

	-- we should be able to migrate saved data formats immediately
	local DBVERSION = 3
	local dbchar = self.db.char
	local dbglob = self.db.global
	
	dbchar.version = dbchar.version or DBVERSION
	dbglob.version = dbglob.version or DBVERSION

	if dbchar.version == DBVERSION and dbglob.version == DBVERSION then
		return
	end

	if dbchar.version < 2 then
		-- version 1 stored a map of spec ids by boss names under char
		-- version 2 replaces this with a map of encounter ids by boss names under global
		-- spec ids can then be found with the existing spec ids by encounter ids map under char
		for encounter, spec in pairs(dbchar.encounters) do
			for _, name in EncounterBosses, encounter do
				if dbchar.bosses[name] ~= spec then
					if not dbchar.bosses[name] then
						dbgprint.error(format("migration error: encounter %i missing boss %s", encounter, name))
					else
						dbgprint.error(format("migration error: encounter %i set to %i but boss %s set to %i", encounter, spec, name, dbchar.bosses[name]))
					end
				end
				dbglob.bosses[name] = encounter
				dbchar.bosses[name] = nil
			end
		end
		dbchar.bosses = nil

		-- putting both values into a subtable
		if dbchar.savedLootSpec then 
			dbchar.savedLoot = { spec = dbchar.savedLootSpec, boss = dbchar.savedLootBoss }
		end
		dbchar.savedLootSpec = nil
		dbchar.savedLootBoss = nil

		-- bosses seen was just some debug code I had, shouldn't be in users settings
		dbglob.bossesSeen = nil
	end

	if dbchar.version < 3 then
		-- loot spec choices were saved in a table indexed by encounter id
		-- version 3 stores them by encounter id and then by difficulty id
		-- for now going to punt on migration (sorry alpha users) and start fresh
		wipe(dbchar.encounters)
	end

	dbchar.version = DBVERSION
	dbglob.version = DBVERSION
end


-- Called when the addon is enabled
function Otis:OnEnable()
	self:LoadSavedSettings()
	dbchar = self.db.char
	dbbosses = self.db.global.bosses
	dbworld  = self.db.global.worldBosses

	self:SetDebugLevel(DL[self.db.global.output])

	g_idPlayerClass = select(3, UnitClass("player"))
	for i=1, GetNumSpecializationsForClassID(g_idPlayerClass) do
		local id, name = GetSpecializationInfo(i)
		g_tblSpecNames[id] = name
	end


	self:RegisterEvent("PLAYER_LOOT_SPEC_UPDATED")

	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")

	self:RegisterEvent("PLAYER_DEAD")
--	self:RegisterEvent("PLAYER_ALIVE")
--	self:RegisterEvent("PLAYER_UNGHOST")
--	self:RegisterEvent("PLAYER_ENTERING_WORLD")

	self:SecureHookScript(BonusRollFrame, "OnShow", "WaitForBonusLoot")
	self:SecureHookScript(BonusRollFrame, "OnHide", "WaitFinished")
	self:RegisterEvent("MISSING_OUT_ON_LOOT")

	self:RegisterEvent("EJ_LOOT_DATA_RECIEVED", "MakeButtons")
	self:RegisterEvent("EJ_DIFFICULTY_UPDATE",  "MakeButtons")
end

-- Called when the addon is disabled
function Otis:OnDisable()
	self:UnregisterEvent("PLAYER_LOOT_SPEC_UPDATED")

	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")

	self:UnregisterEvent("PLAYER_DEAD")
--	self:UnregisterEvent("PLAYER_ALIVE")
--	self:UnregisterEvent("PLAYER_UNGHOST")
--	self:UnregisterEvent("PLAYER_ENTERING_WORLD")

	self:UnregisterEvent("EJ_LOOT_DATA_RECIEVED")
	self:UnregisterEvent("EJ_DIFFICULTY_UPDATE")

	self:UnregisterEvent("MISSING_OUT_ON_LOOT")

	self:UnhookAll()
end

function Otis:PLAYER_REGEN_DISABLED() -- entered combat
	dbgprint.events("PLAYER_REGEN_DISABLED - combat started")
	
	-- we could get pulled into a combat before the restore timer from the previous combat has fired
	TimerRestore:Cancel()

	if dbchar.savedLoot then
		-- we could be reentering combat with boss after a resurrection or after a disconnect
		-- or someone could also have started a new fight while we're still deciding whether 
		-- to roll in the bonus loot dialog
		if g_fWaitingOnBonusLoot then
			-- if the bonus loot roll window really is still up then we really don't want to do 
			-- anything with it until the user is done. we should exit without setting subscribing
			-- to any events. we can retry things when the loot window closes
			if BonusRollFrame and BonusRollFrame:IsVisible() then
				dbgprint.config(format("loot window for %s still up; ignoring start of new combat until resolved", dbchar.savedLoot.boss or "ERROR"))
				return
			else
				dbgprint.events(format("player presumed to have died fighting %s; do not wait on bonus roll window"), dbchar.savedLoot.boss or "ERROR")
				g_fWaitingOnBonusLoot = false
			end
		end

		dbgprint.config(format("player may have wiped or reentered combat with %s after resurrection or disconnect. restore, clear and rerun boss detection", dbchar.savedLoot.boss or "ERROR"))
		self:RestoreLootSpec()
	end

	if IsInInstance() then
		g_idActiveDifficulty = select(3, GetInstanceInfo())

		self:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")

		-- frames can sometimes go up before we enter combat, 
		-- especially when stealthed. force a check just in case
		self:INSTANCE_ENCOUNTER_ENGAGE_UNIT("INSTANCE_ENCOUNTER_ENGAGE_UNIT - forced at combat start")
	else
		g_idActiveDifficulty = 0 -- assume world boss

--		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
		self:RegisterEvent("UNIT_CLASSIFICATION_CHANGED")

		-- They may never actually CHANGE target in the fight if they start
		-- out targeted on the boss. So, call the handler at the start.
		self:PLAYER_TARGET_CHANGED("PLAYER_TARGET_CHANGED - forced at combat start")
	end
end

function Otis:INSTANCE_ENCOUNTER_ENGAGE_UNIT(...)
	dbgprint.events(...)
	
	for i=1, 5 do
		local unit = "boss"..i
		local name = UnitName(unit)
		if name then
--[===[@debug@
			self.db.global.bossesSeen[name] = true
--@end-debug@]===]
			dbgprint.test(unit,name)

			local encounter = dbbosses[name]
			if encounter then
				local spec = dbchar.encounters[encounter][g_idActiveDifficulty]
				if spec then
					self:ChangeLootSpec(spec, name, unit)
				end

				-- we've identified this encounter, don't need to process more boss frame updates
				self:UnregisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
				return
			end
		end
	end
end

function Otis:PLAYER_TARGET_CHANGED()
	self:CheckUnit("target")
end

function Otis:UPDATE_MOUSEOVER_UNIT()
	self:CheckUnit("mouseover")
end

function Otis:UNIT_CLASSIFICATION_CHANGED(event, unit)
	-- surely this is a rare event
	dbgprint.events(event, unit)
	self:CheckUnit(unit)
end

function Otis:CheckUnit(unit)
	if not UnitExists(unit) then return end
	
	if UnitIsPlayer(unit) then
		unit = unit .. "target"
		if not UnitExists(unit) then return end
	end

	-- the stupid celestials don't appear to be worldbosses ... how lame
	local classification = UnitClassification(unit)
	if  classification == "worldboss" or classification == "elite" then
		-- UnitIsTappedByPlayer(unit) -- at the moment going to say that we don't really care if the unit has been tapped by the player or not because we see no real harm
		local name = UnitName(unit)
		local encounter = dbbosses[name]
		if encounter then
			local spec = dbchar.encounters[encounter][g_idActiveDifficulty]
			if spec then
				-- dbgprint.test("CheckUnit would have changed loot spec here for", unit, name)
				self:ChangeLootSpec(spec, name, unit)
			end

			-- we've identified this encounter, don't need to process more target updates
			self:UnregisterEvent("PLAYER_TARGET_CHANGED")
			self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
			self:UnregisterEvent("UNIT_CLASSIFICATION_CHANGED")
		end
	end
end

function Otis:COMBAT_LOG_EVENT_UNFILTERED(event, _, _, _, _, srcName, _, _, _, dstName)
	-- local specID = dbbosses[srcName] or dbbosses[dstName]
	local name, encounter = srcName, dbbosses[srcName]
	if not encounter then
		name, encounter = dstName, dbbosses[dstName]
	end

	if encounter then
		local spec = dbchar.encounters[encounter][g_idActiveDifficulty]
		if spec then
			self:ChangeLootSpec(spec, name, "")
		end

		-- we've identified this encounter, don't need to process more combat log notifications
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

function Otis:PLAYER_REGEN_ENABLED(event) -- left combat
	dbgprint.events("PLAYER_REGEN_ENABLED - combat ended")
	-- self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:UnregisterEvent("UNIT_CLASSIFICATION_CHANGED")
	self:UnregisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")

	g_idActiveDifficulty = nil

	if dbchar.savedLoot and not g_fWaitingOnBonusLoot then
		TimerRestore:Set(event)
	end
end

function Otis:PLAYER_LOOT_SPEC_UPDATED(event)
	local newspec = GetLootSpecialization()
	if not g_idPendingSpec then
		dbgprint.events(format("PLAYER_LOOT_SPEC_UPDATED to %s - user initiated", g_tblSpecNames[newspec]))
		if dbchar.savedLoot then
			dbgprint.config("removing restore point due to user initiated spec change")
			dbchar.savedLoot = nil
		end
	elseif newspec == g_idPendingSpec then
		dbgprint.events(format("PLAYER_LOOT_SPEC_UPDATED to %s - expected by Otis", g_tblSpecNames[newspec]))
	else
		dbgprint.error(format("ERROR: PLAYER_LOOT_SPEC_UPDATED to %s, expected %s", g_tblSpecNames[newspec], g_tblSpecNames[g_idPendingSpec]))
	end
	g_idPendingSpec = nil	
end

function Otis:ChangeLootSpec(spec, name, unit)
	dbgprint.config("ChangeLootSpec", spec, name, unit)
	local current = GetLootSpecialization()

	-- g_idPendingSpec should never equal current and if it does we should
	-- probably assume the game won't send an event and therefore its the 
	-- same as if g_idPendingSpec wasn't set at all. No-op changes below
	-- should be ensure the value is nil.

	if g_idPendingSpec and g_idPendingSpec ~= current then
		-- g_idPendingSpec will exist when you wipe on a boss we set loot for, 
		-- enter a new combat and restore to the saved spec, then recognize a 
		-- boss in the new combat and come here to set its spec. The new set
		-- can happen before the restore finishes propagating through the game.
		if spec == g_idPendingSpec then
			dbgprint.config(format("a change to %s is already pending", g_tblSpecNames[spec]))
			g_idPendingSpec = nil
			SetLootSpecialization(spec) -- this should be an unnecessary no-op, but do it just in case ...
		else
			-- we're assuming the pending spec is from a restore; therefore the value we
			-- revert to after the current combat should be the same spec as is pending
			dbchar.savedLoot = { spec = g_idPendingSpec, boss = name }

			-- if spec == current we will not get a notification and so should not set a new value for pending
			-- the no-op does prevent the pending from happening though, and so must be done
			if spec == current then
				dbgprint.config(format("reverting spec change from %s to %s for %s %s", g_tblSpecNames[spec], g_tblSpecNames[g_idPendingSpec], unit, name))
				g_idPendingSpec = nil
				SetLootSpecialization(spec)
			else
				dbgprint.config(format("changing loot spec from %s to %s for %s %s", g_tblSpecNames[g_idPendingSpec], g_tblSpecNames[spec], unit, name))
				g_idPendingSpec = spec
				SetLootSpecialization(spec)
			end
		end
	elseif current ~= spec then
		dbchar.savedLoot = { spec = current, boss = name }

		dbgprint.config(format("changing loot spec from %s to %s for %s %s", g_tblSpecNames[current], g_tblSpecNames[spec], unit, name))
		g_idPendingSpec = spec
		SetLootSpecialization(spec)
	else  -- current == spec
		dbgprint.config(format("current loot spec already matches %s for %s %s", g_tblSpecNames[spec], unit, name))
		g_idPendingSpec = nil
	end
end

function Otis:RestoreLootSpec()
	if g_fWaitingOnBonusLoot then
		dbgprint.test("ERROR - restore called while we're supposed to be waiting on bonus loot")
		g_fWaitingOnBonusLoot = false
	end

	local savedLoot = dbchar.savedLoot
	if savedLoot then
		-- g_idPendingSpec should never be set when we get to a restore, right?
		if g_idPendingSpec then
			dbgprint.test(format("possible error: pending spec is set to %s", g_tblSpecNames[g_idPendingSpec]))
		end

		local current = GetLootSpecialization()
		if current == savedLoot.spec then
			dbgprint.config(format("current loot spec %s already matches the spec we overrode for %s", g_tblSpecNames[current], savedLoot.boss))
			g_idPendingSpec = nil
			SetLootSpecialization(savedLoot.spec) -- this should be an unnecessary no-op, but do it just in case ...
		else
			dbgprint.config(format("restoring loot spec to %s after %s fight", g_tblSpecNames[savedLoot.spec], savedLoot.boss))
			g_idPendingSpec = savedLoot.spec
			SetLootSpecialization(savedLoot.spec)
		end

		dbchar.savedLoot = nil
	end
end

function Otis:WaitForBonusLoot()
	-- if we didn't change spec, then we don't have to wait. this boss means nothing to us
	if dbchar.savedLoot then
		dbgprint.events("waiting for bonus loot")

		TimerRestore:Cancel()
		g_fWaitingOnBonusLoot = true
	end
end

function Otis:WaitFinished()
	g_fWaitingOnBonusLoot = false
	if InCombatLockdown() then
		-- if we enter combat while waiting on bonus loot window, then we won't
		-- try to identify the encounter or do anything until resolved. now that
		-- it has, pretend that combat has just started
		self:PLAYER_REGEN_DISABLED()
	else
		TimerRestore:Set("BonusRollFrame_OnHide")
	end
end

function Otis:MISSING_OUT_ON_LOOT(event, ...)
	dbgprint.test(event, ...)
	g_fWaitingOnBonusLoot = false
	TimerRestore:Set(event)
end

function Otis:PLAYER_DEAD(event)
	dbgprint.spam(event)
	self:WaitForBonusLoot()
end

function Otis:PLAYER_ALIVE(event)
	dbgprint.spam(event)
	-- against a world boss, you could rez and have the boss die before you 
	-- reentered combat. don't touch anything until next combat
	-- self:WaitFinished()
end

function Otis:PLAYER_UNGHOST(event)
	dbgprint.spam(event)
	-- against a world boss, you could rez and have the boss die before you 
	-- reentered combat. don't touch anything until next combat
	-- self:WaitFinished()
end

function Otis:state()
	local current = g_tblSpecNames[GetLootSpecialization()]
	local savedLoot = dbchar.savedLoot
	local savedBoss = dbchar.savedLoot and savedLoot.boss or "nil"
	local savedSpec = dbchar.savedLoot and g_tblSpecNames[dbchar.savedLoot.spec] or "nil"
	local pending = g_idPendingSpec and g_tblSpecNames[g_idPendingSpec] or "nil"
	self:Print(format("current: %s  savedBoss: %s  savedSpec: %s  waiting: %s  pending: %s", current, savedBoss, savedSpec, tostring(g_fWaitingOnBonusLoot), pending))
end
